import sys, os.path
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..'))

import weakref
import gtk
import gc

import unittest

from gsignals import Signal, SignalManager


def refresh_gui():
    """
    Magic function to emulate gtk main loop
    """
    while gtk.events_pending():
        gtk.main_iteration_do(block=False)
  

class Signals(SignalManager):
    test1 = Signal()
    test2 = Signal()
    

class SignalChecker(object):
    def __init__(self):
        self.signals = {}
        
    def clear(self):
        self.signals.clear()
        
    def fired(self, name):
        return name in self.signals            
    
    def fire(self, name):
        self.signals[name] = True


class Plugin(object):
    
    def __init__(self, signals, checker):
        self.signals = signals
        signals.connect_signals(self)
        self.checker = checker
    
    @Signals.test1
    def test1(self, *args):
        self.checker.fire('test1')
    
    @Signals.test2
    def test2(self, *args):
        self.checker.fire('test2')
        
        
class TestCase(unittest.TestCase):
    
    def test_plugin_deletion(self):
        checker = SignalChecker()
        
        # This trick allows to brake direct refs from current stack frame
        signal_holder = [Signals()]
        wsignals = weakref.ref(signal_holder[0]) # to check signal manager deletion
        wsender = weakref.ref(wsignals().sender) # to check signal manager GObject deletion
        
        plugin_holder = [Plugin(wsignals(), checker)] 
        wplugin = weakref.ref(plugin_holder[0]) # to check plugin deletion
        
        test1_handler_id = wplugin().test1.handler.id
        test2_handler_id = wplugin().test2.handler.id
        
        # There is can be only one ref from holder list
        plugin_references = gc.get_referrers(wplugin())
        self.assertEquals(len(plugin_references), 1)
        self.assertTrue(plugin_references[0] is plugin_holder)
        
        wsignals().test1.emit()
        refresh_gui()
        self.assertTrue(checker.fired('test1'))
        
        # Now we unload plugin by eleminating direct refs.
        # In our case simply clear holder list
        plugin_holder[:] = []
        checker.clear()
        
        self.assertEqual(wplugin(), None) # plugin really deleted
        
        # At this point we have Signals GObject connected to plugin callbacks
        self.assertTrue(wsignals().sender.handler_is_connected(test1_handler_id))
        self.assertTrue(wsignals().sender.handler_is_connected(test2_handler_id))
        
        # But after emmiting signal it is automatically disconnected
        wsignals().test2.emit()
        refresh_gui()
        self.assertFalse(checker.fired('test2'))
        self.assertTrue(wsender().handler_is_connected(test1_handler_id))
        self.assertFalse(wsender().handler_is_connected(test2_handler_id))
        
        # Removing refs to signal manager. Manager and its GObject must vanish 
        signal_holder[:] = []
        self.assertEqual(wsignals(), None) # Signal manager really deleted
        self.assertEqual(wsender(), None) # GObject really deleted
        
        
if __name__ == '__main__':
    unittest.main()  



import sys
import unittest

import gtk
import gobject

import sys, os
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..'))

from gsignals import Signal, SignalManager, connect_external, connect_all
from gsignals.signals import SSIGNAL


def refresh_gui():
    """
    Magic function to emulate gtk main loop
    """
    while gtk.events_pending():
        gtk.main_iteration_do(block=False)


class Case(unittest.TestCase):
    
    def test_signal_simple_creation(self):
        s = Signal()
        
        self.assertEqual(s.signal_type, SSIGNAL) 
        self.assertEqual(s.return_type, None)
        self.assertEqual(s.arg_types, ())

    def test_signal_creation_with_arguments(self):
        s = Signal(object, int)
        
        self.assertEqual(s.signal_type, SSIGNAL) 
        self.assertEqual(s.return_type, None)
        self.assertEqual(s.arg_types, (object, int))

    def test_signal_creation_with_return_type(self):
        s = Signal(return_type=object)
        
        self.assertEqual(s.signal_type, SSIGNAL) 
        self.assertEqual(s.return_type, object)
        self.assertEqual(s.arg_types, ())

    def test_signal_creation_with_signal_type(self):
        s = Signal(type=gobject.SIGNAL_ACTION)
        
        self.assertEqual(s.signal_type, gobject.SIGNAL_ACTION) 
        self.assertEqual(s.return_type, None)
        self.assertEqual(s.arg_types, ())
        
    def test_simple_autoconnecting_and_emmiting_case(self):
        
        class Signals(SignalManager):
            hide = Signal(int)
            show = Signal()
            
        class Handler(object):
            def __init__(self):
                self.signals = Signals()
                connect_all(self, self.signals)
                
                self.hided = None
                self.showed = None
            
            @Signals.hide(idle=True)
            def hide(self, sender, arg):
                self.hided = arg
            
            @Signals.show
            def show(self, *args):
                self.showed = True
            
        h = Handler()
        
        h.signals.show.emit()
        self.assertTrue(h.showed)
        
        h.signals.hide.emit(5)
        # callback will be called after event loop processing because idle wrapper
        self.assertEqual(h.hided, None)
        refresh_gui()
        self.assertEqual(h.hided, 5)
        
    def test_autoconnecting_to_external_signals(self):
        class Signals1(SignalManager):
            fire = Signal()
            
        class Signals2(SignalManager):
            fire = Signal()
        
        class Handler(object):
            def __init__(self, s1, s2):
                self.fired = 0
                connect_all(self, sender1=s1, sender2=s2)
            
            @connect_external('sender1', 'fire')
            @connect_external('sender2', 'fire', idle=True)
            def fire(self, *args):
                self.fired += 1

        s1 = Signals1()
        s2 = Signals2()
        h = Handler(s1.sender, s2.sender)
        
        s1.fire.emit()
        self.assertEqual(h.fired, 1)
        
        s2.fire.emit()
        self.assertEqual(h.fired, 1)
        refresh_gui()
        self.assertEqual(h.fired, 2)
        
    def test_signals_blocking(self):
        class Signals(SignalManager):
            fire = Signal()
            
        class Handler(object):
            def __init__(self):
                self.fired = 0
                self.signals = Signals()
                self.signals.connect_signals(self)
            
            @Signals.fire
            def fire(self, *args):
                self.fired += 1

        h = Handler()

        h.fire.handler.block()
        h.signals.fire.emit()
        self.assertEqual(h.fired, 0)
        
        h.fire.handler.unblock()
        h.signals.fire.emit()
        self.assertEqual(h.fired, 1)

    def test_several_signals_blocking(self):
        class FireSignals(SignalManager):
            fire = Signal()

        class WaterSignals(SignalManager):
            water = Signal()
            
        class Handler(object):
            def __init__(self, water):
                self.fired = 0
                self.signals = FireSignals()
                
                connect_all(self, self.signals, water=water.sender)
            
            @FireSignals.fire
            @connect_external('water', 'water')
            def fire(self, *args):
                self.fired += 1

        w = WaterSignals()
        h = Handler(w)

        self.assertRaises(Exception, h.fire.handler.block)
        
        h.fire.handler(signal_name='fire').block()
        w.water.emit()
        h.signals.fire.emit()
        self.assertEqual(h.fired, 1)
        
        h.fire.handler(signal_name='fire').unblock()
        w.water.emit()
        h.signals.fire.emit()
        self.assertEqual(h.fired, 3)
        
        h.fire.handler(sender_name='water').block()
        w.water.emit()
        h.signals.fire.emit()
        self.assertEqual(h.fired, 4)
        
        h.fire.handler(sender_name='water').unblock()
        w.water.emit()
        h.signals.fire.emit()
        self.assertEqual(h.fired, 6)
        
        h.fire.handler(sender=w.sender).block()
        w.water.emit()
        h.signals.fire.emit()
        self.assertEqual(h.fired, 7)        



import weakref
from gobject import idle_add

class WeakCallback(object):
    """
    Weak callback functor which disconnects on real callback deletion
    
    Magic goes here. It allows to skip accurate manual object disposion by brokening
    cyclic reference beetween sender and callback. Let python gc to work.
    
    Also it can wrap callback in idle_add with specified priority.
    
    Important note! Object and its callback attribute are passed separately.
    Because bounded methods are too weak.
    """
    def __init__(self, obj, attr, idle):
        self.wref = weakref.ref(obj)
        self.callback_attr = attr
        self.gobject_token = None
        self.dbus_token = None
        self.idle = idle

    def __call__(self, *args, **kwargs):
        obj = self.wref()
        if obj:
            # object still alive so calling method 
            attr = getattr(obj, self.callback_attr)
            
            if self.idle is False or self.idle is None:
                return attr(*args, **kwargs)
            elif self.idle is True:
                idle_add(attr, *args, **kwargs)
            else:
                idle_add(attr, priority=self.idle, *args, **kwargs)
                    
        elif self.gobject_token:
            sender = args[0]
            sender.disconnect(self.gobject_token)
            self.gobject_token = None
        elif self.dbus_token:
            self.dbus_token.remove()
            self.dbus_token = None
            
        return False


def weak_connect(sender, signal, connector, attr, idle=False, after=False):
    """
    Function to connect some GObject with weak callback
    """
    wc = WeakCallback(connector, attr, idle)
    
    if after:
        wc.gobject_token = sender.connect_after(signal, wc)
    else:
        wc.gobject_token = sender.connect(signal, wc)

    #print "Connected", sender, signal, connector, attr, idle, after

    return wc.gobject_token



import gobject
import gobject.constants
import weakref

from weak import weak_connect
from util import append_attr

SSIGNAL = gobject.SIGNAL_RUN_LAST | gobject.SIGNAL_NO_RECURSE | gobject.SIGNAL_ACTION
SACTION = gobject.SIGNAL_RUN_LAST | gobject.SIGNAL_ACTION

def attach_signal_connect_info(attr, obj, func, after, idle):
    """
    Adds signal connection info to function
    
    Used by signal and trigger decorators
    """
    connect_params = dict(after=after, idle=idle)

    if func:
        if not getattr(func, '__call__'):
            raise Exception('Signal decorator accept callable or connect params')
        
        append_attr(func, attr, (obj, connect_params))
        return func
    else:
        def inner(func):
            append_attr(func, attr, (obj, connect_params))
            return func
        
        return inner    
    
class Signal(object):
    """
    Unbounded signal
    
    Class holds signal parameters which used to construct correct GObject later.
    
    Instantiating signals::    
        
        Signal() # Signal without arguments
        
        Signal(object, int) # Signal with two arguments
        
        Signal(object, return_type=int) # Signal with return type
        
        Signal(type=gobject.SIGNAL_RUN_FIRST) # default signal type is gobject.SIGNAL_RUN_LAST | gobject.SIGNAL_NO_RECURSE | gobject.SIGNAL_ACTION 


    Unbounded signal instances can be used to mark callbacks for automatic signal connecting::        

        signal = Signal()
        
        class Handler(object):
            @signal
            def callback(...): pass # Usual (in gobject terms) signal connection
            
            @signal(idle=True)
            def callback(...): pass # Connects signal with idle wrapper
            
            @signal(after=True)
            def callback(...): pass # sender.connect_after(callback) analog
            
            @signal(idle=9999)
            def callback(...): pass # idle wrapper will start callback with specified priority
    """
    
    def __init__(self, *signal_args, **kwargs):
        allowed_named_arguments = set(('type', 'return_type'))
        if not all(r in allowed_named_arguments for r in kwargs.keys()): 
            raise Exception('Signal constructor takes only `type` and `return_type` named arguments')
        
        self.signal_type = kwargs.get('type', SSIGNAL)
        self.return_type = kwargs.get('return_type', None)
        self.arg_types = tuple(signal_args)
        self.name = None

    def __call__(self, func=None, after=False, idle=False):
        return attach_signal_connect_info('signals_to_connect', self, func, after, idle)
            
    def emit(self):
        """
        Only hint for IDE
        """
        raise Exception('You cannot emit unbounded signals')


class SignalManager(object):
    """
    Wrapper for inner GObject with signals
    
    Example::
        
        class Manager(SignalManager):
            show = Signal()
            hide = Signal()
        
    ``Manager.show`` and ``Manager.hide`` is unbounded signals and can be used as
    decorators to callbacks. Whereas ``instance.show`` and ``instance.hide`` is bounded and
    can be used to emit signals:: 
    
        class Plugin(object):
            def __init__(self):
                self.signals = Manager()
                self.signals.connect_signals()
                
                self.signals.hide.emit() 
                
            @Manager.show
            def show(self, sender):
                pass
                
    Inner GObject with necessary __gsignals__ is constructed during instance initialization
    """
    def __init__(self):    
        signals = {}
        for sname, signal in self.__class__.__dict__.iteritems():
            if isinstance(signal, Signal):
                signal.name = sname.replace('_', '-')
                signals[signal.name] = (signal.signal_type,
                    signal.return_type, signal.arg_types)
                
                setattr(self, sname, BoundedSignal(self, signal))
        
        if signals:
            classname = self.__class__.__name__ + 'Signals'
            cls = type(classname, (gobject.GObject,), {'__gsignals__':signals})
            gobject.type_register(cls)
            self.sender = cls()
        else:
            raise Exception('Empty signal manager')
            
    def connect_signals(self, obj):
        """
        Connects marked object methods
        """
        for attr, value in obj.__class__.__dict__.iteritems():
            for signal, connect_params in getattr(value, 'signals_to_connect', ()):
                id = self.connect(signal, obj, attr, **connect_params)
                append_handler_to_callback(value, id, self.sender, signal.name)    

    def connect(self, signal, obj, attr, after, idle):
        """
        Connects unbounded signal
        
        @param signal: Unbounded signal
        """
        return weak_connect(self.sender, signal.name, obj, attr, after=after, idle=idle)


class BoundedSignal(object):
    """
    This class knows about its GObject wrapper and unbounded signal name
    
    This allows it to emit signals. Bounded signal weakly connected to its manager so
    you can safely use it in any context 
    """
    def __init__(self, manager, signal):
        self.manager = weakref.ref(manager)
        self.signal = signal

    def connect(self, obj, attr, after, idle):
        manager = self.manager()
        if manager: 
            manager.connect(self.signal, obj, attr, after=after, idle=idle)

    def emit(self, *args):
        manager = self.manager()
        if manager: 
            manager.sender.emit(self.signal.name, *args)


def connect_external(sender_name, signal_name, after=False, idle=False):
    def inner(func):
        return attach_signal_connect_info('external_signals_to_connect',
            (sender_name, signal_name), func, after, idle)
        
    return inner

def connect_external_signals(obj, **kwargs):
    for attr, value in obj.__class__.__dict__.iteritems():
        for (sender_name, signal_name), connect_params in getattr(value, 'external_signals_to_connect', ()):
            sender = kwargs[sender_name]
            id = weak_connect(sender, signal_name, obj, attr, **connect_params)
            append_handler_to_callback(value, id, sender, signal_name, sender_name)
            
def append_handler_to_callback(callback, handler_id, sender, signal_name, sender_name=None):
    if not hasattr(callback, 'handler'):
        callback.handler = HandlerHolder()
        
    callback.handler.add(handler_id, sender, signal_name, sender_name)

def connect_all(obj, *signal_managers, **external_senders): 
    [s.connect_signals(obj) for s in signal_managers]
    
    if external_senders:
        connect_external_signals(obj, **external_senders)    

class Handler(object):
    def __init__(self, handler_id, sender, signal_name, sender_name):
        self.id = handler_id
        self.sender = weakref.ref(sender)
        self.signal_name = signal_name
        self.sender_name = sender_name
        
    def is_match(self, sender, sender_name, signal_name):
        result = True
        
        result = result and (sender == None or self.sender() is sender)
        result = result and (sender_name == None or self.sender_name == sender_name)
        result = result and (signal_name == None or self.signal_name == signal_name)
        
        return result
        
    def block(self):
        sender = self.sender()
        if sender:
            sender.handler_block(self.id)
        
    def unblock(self):
        sender = self.sender()
        if sender:
            sender.handler_unblock(self.id)


class HandlerHolder(object):

    def __init__(self):
        self.handlers = []
        
    def add(self, id, sender, signal_name, sender_name=None): 
        self.handlers.append(Handler(id, sender, signal_name, sender_name))

    def block(self):
        try:
            (handler,) = self.handlers
            handler.block()
        except ValueError:
            raise Exception('There are several signals connected to callback')
        
    def unblock(self):
        try:
            (handler,) = self.handlers
            handler.unblock()
        except ValueError:
            raise Exception('There are several signals connected to callback')

    @property
    def id(self):
        try:
            (handler,) = self.handlers
            return handler.id
        except ValueError:
            raise Exception('There are several signals connected to callback')

    def __call__(self, sender=None, sender_name=None, signal_name=None):
        handler = None
        for h in self.handlers:
            if h.is_match(sender=sender, sender_name=sender_name, signal_name=signal_name):
                if handler:
                    raise Exception('Match returns several handlers')
                else:
                    handler = h
                    
        return handler



from setuptools import setup, find_packages

def run_tests():
    import unittest, sys, os.path, glob
    sys.path.insert(0, os.path.join(os.path.dirname(__file__), 'tests'))
    return unittest.defaultTestLoader.loadTestsFromNames(
        map(lambda r: os.path.splitext(os.path.basename(r))[0], glob.glob('tests/test_*.py')))

setup(
    name     = 'gsignals',
    version  = '0.1',
    author   = 'Anton Bobrov',
    author_email = 'bobrov@vl.ru',
    packages     = find_packages(),
    description = 'Wrappers for gobject signals',
    #long_description = open('README.txt').read(),
    zip_safe   = False,
    test_suite = '__main__.run_tests',
    url = 'http://github.com/baverman/scribes-goodies',
)



def append_attr(obj, attr, value):
    """
    Appends value to object attribute
    
    Attribute may be undefined
    
    For example:
        append_attr(obj, 'test', 1)
        append_attr(obj, 'test', 2)
        
        assert obj.test == [1, 2]
    """
    try:
        getattr(obj, attr).append(value)
    except AttributeError:
        setattr(obj, attr, [value])
